import time
import chess


class TimeControl(object):
    """Class modeling an arbitrary time control.

    Starting time and increment are given in units of seconds,
    and then this class is responsible for the handling of each
    players time.

    The end user _never_ needs to use this class any further than
    instantiating an object with the desired time controls and providing
    this instance to the functions that expect it.
    """

    def __init__(self, initial_time, increment=0):
        """Initialize a time control with the given starting time and increment.

        Arguments
        ---------
        initial_time: int or float
            Starting time for both players in seconds.
        increment: int or float, optional
            Time in seconds gained after each move.
        """
        self.initial_time = initial_time
        self.increment = increment
        self.wtime = self.btime = initial_time

    def _start_new_game(self, side_to_move=chess.WHITE):
        """Reset times and set side_to_move."""
        self.wtime = self.btime = self.initial_time
        self.side_to_move = side_to_move
        self.timepoint = time.time()

    def _signal_move_made(self):
        """Update times since last call and side_to_move."""
        new_time = time.time()
        self.timepoint, delta = new_time, new_time - self.timepoint

        if self.side_to_move == chess.WHITE:
            self.wtime = self.wtime - delta + self.increment
        else:
            self.btime = self.btime - delta + self.increment

        self.side_to_move = not self.side_to_move

        return self.wtime <= 0 or self.btime <= 0



import chess
import chess.polyglot
import random


def play_from_opening_book(
    book, max_depth=10, fen=chess.STARTING_FEN, random_seed=None
):
    """Play out moves from an opening book and return the resulting board.

    From the given `fen` starting position, draw weighted random moves from the opening book
    to a maximum depth of `2 * max_depth` plies. Whenever there are no move moves in the book,
    the play stops and the board returned.

    If a seed integer is given, then this will always return the same final position.

    Arguments
    ---------
    book: str
        Path to a polyglot opening book file.
    max_depth: int, optional
        The maximum depth to play to. The number of moves (plies) made will at most be 2 times this.
        Default is 10.
    fen: str, optional
        Starting position in FEN notation. Default is the standard opening position.
    random_seed: int, optional
        Seed the random number generator to produce the same results each call. Default is to not seed,
        and so successive calls will in general yield different boards.

    Returns
    -------
        A `chess.Board` with the resulting position.
    """
    if random_seed is not None:
        random.seed(random_seed)

    board = chess.Board(fen)

    with chess.polyglot.MemoryMappedReader(book) as reader:
        try:
            for _ in range(2 * max_depth):
                move = reader.weighted_choice(board).move()
                board.push(move)
        except IndexError:
            pass

    return board



import chess
from itertools import permutations
from .match import play_match
from .openingbook import play_from_opening_book


def _get_match_id(p1, p2, round_id):
    """Return a unique integer for each pair of p1 and p2 for every value of round_id.

    Property:
        _get_match_id(p1, p2, r) == _get_match_id(p2, p1, r) for all p1, p2, r
    """
    return hash("".join(sorted(str(p1) + str(p2) + str(round_id))))


def play_tournament(
    players,
    time_control,
    n_games=1,
    opening_book=None,
    opening_book_depth=10,
    repeat=True,
    player_options=dict(),
):
    """Play all possible matches between the given players.

    Each possible matchup, including white/black permutations, will be played `n_games` times each.

    Arguments
    ---------
    players: list of str
        A list of executables of the UCI engines to play.
    time_control: chester.timecontrol.TimeControl
        A TimeControl instance describing the time control to use.
    n_games: int
        The number of times each match-up should be played.
    opening_book: str, optional
        If the openings should be selected from an opening book, give the
        path to a polyglot opening book as this parameter. Default is to not
        use a book.
    opening_book_depth: int, optional
        Maximum depth to play out from the `opening_book`. Has no effect if no book is given.
        Default is a depth of 10 (20 plies).
    repeat: bool, optional
        When using an opening book, setting this to True will ensure that
        each player plays both sides of the same opening vs the same opponent.
        Default value is True.
    player_options: dict of dict, optional
        A dictionary of option dictionaries corresponding to a player. The dict
        may be empty, and should be indexable by the given string in `players`.
        See `chester.match.play_match` for details on the format of each dictionary.

        Example:

        ``` python
        player_options = {"stockfish": {"Contempt": 24, "MultiPV": 3}}
        ```

    Returns
    -------
        Generator of `chess.pgn.Game` objects, exactly `n_games * len(list(permutations(players, 2)))` long.
    """
    for round_count in range(1, n_games + 1):
        for white, black in permutations(players, 2):
            if opening_book:
                if repeat:
                    board = play_from_opening_book(
                        opening_book,
                        max_depth=opening_book_depth,
                        random_seed=_get_match_id(white, black, round_count),
                    )
                else:
                    board = play_from_opening_book(
                        opening_book, max_depth=opening_book_depth
                    )
            else:
                board = chess.Board()

            pgn = play_match(
                white,
                black,
                time_control,
                fen=board.fen(),
                engine1_options=player_options.get(white, dict()),
                engine2_options=player_options.get(black, dict()),
            )
            pgn.headers["Round"] = round_count
            yield pgn



import chess
import chess.engine
import chess.pgn
import datetime


def play_match(
    engine1,
    engine2,
    time_control,
    fen=chess.STARTING_FEN,
    engine1_options=dict(),
    engine2_options=dict(),
):
    """Play a match between the given engines with a given time control.

    Arguments
    ---------
    engine1, engine2: str
        Executable of the UCI engines to play. Engine1 will play white.
    time_control: chester.timecontrol.TimeControl
        A TimeControl instance describing the time control to use.
    fen: str, optional
        Starting position in FEN notation. Default is the standard opening position.
    engine1_options, engine2_options: dict, optional
        A dictionary with options to set for each engine. The key should be the
        name of the option as provided by the engine. To see available options you may
        do:
        ``` python
        engine = chess.engine.SimpleEngine.popen_uci("name-of-executable")
        print(engine.options)
        ```
        See https://python-chess.readthedocs.io/en/latest/engine.html#chess.engine.EngineProtocol.options

        Only the options that match the engines option names will be configured.

    Returns
    -------
        A `chess.pgn.Game` instance describing the game.
    """
    engine1 = chess.engine.SimpleEngine.popen_uci(engine1)
    engine2 = chess.engine.SimpleEngine.popen_uci(engine2)

    # Apply
    for engine, options in zip([engine1, engine2], [engine1_options, engine2_options]):
        available = engine1.options
        options = {
            k: v
            for k, v in options.items()
            if k in available and not available[k].is_managed()
        }
        engine.configure(options)

    result = None
    board = chess.Board(fen)
    time_control._start_new_game(side_to_move=board.turn)
    current_player = engine1 if board.turn == chess.WHITE else engine2
    while not board.is_game_over():
        output = current_player.play(
            board,
            chess.engine.Limit(
                white_clock=time_control.wtime,
                black_clock=time_control.btime,
                white_inc=time_control.increment,
                black_inc=time_control.increment,
            ),
        )

        player_flagged = time_control._signal_move_made()

        if player_flagged:
            if board.has_insufficient_material(not board.turn):
                result = "1/2-1/2"
            else:
                result = "0-1" if board.turn == chess.WHITE else "1-0"
            break

        board.push(output.move)
        current_player = engine1 if board.turn == chess.WHITE else engine2

    pgn = chess.pgn.Game.from_board(board)
    pgn.headers["White"] = engine1.id.get("name")
    pgn.headers["Black"] = engine2.id.get("name")
    pgn.headers["Date"] = datetime.datetime.today().strftime("%Y-%m-%d")
    if fen != chess.STARTING_FEN:
        pgn.headers["FEN"] = fen

    if result is not None:
        pgn.headers["Result"] = result

    engine1.quit()
    engine2.quit()

    return pgn



import os
from setuptools import setup

setup(
    name="chester",
    version="0.3.1",
    author="Bendik Samseth",
    author_email="b.samseth@gmail.com",
    description="Chess Engine Tester - A simple interface to play chess engines against each other, including tournaments.",
    license="MIT",
    keywords="chess engine testing match tournament chester",
    url="https://www.github.com/bsamseth/python-chess-engine-tester",
    packages=["chester"],
    long_description=open(os.path.join(os.path.dirname(__file__), "README.md")).read(),
    long_description_content_type="text/markdown",
    install_requires=["python-chess>=0.26.0"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],
)
